package logwire.web.bo.loader;

import com.google.common.collect.Maps;
import logwire.core.bo.annotation.*;
import logwire.core.bo.field.BizArray;
import logwire.core.bo.field.BizExpando;
import logwire.core.bo.field.BizMany;
import logwire.core.bo.field.BizText;
import logwire.core.bo.object.BizObject;
import logwire.core.exceptions.ApplicationException;
import logwire.core.meta.model.Model;
import logwire.core.meta.model.fields.*;
import logwire.core.meta.query.IQuery;
import logwire.core.resource.BeanLoader;
import logwire.core.resource.loader.JavaModelBeanLoader;
import logwire.core.support.Util;
import logwire.core.tenant.TenantClassLoader;
import logwire.core.tenant.TenantProject;
import logwire.web.bo.field.parse.FieldParser;
import logwire.web.bo.field.parse.FieldParserUtil;
import org.elasticsearch.common.Strings;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.ClassPathScanningCandidateComponentProvider;
import org.springframework.core.io.DefaultResourceLoader;
import org.springframework.core.io.ResourceLoader;
import org.springframework.stereotype.Component;

import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.*;
import java.util.function.Consumer;
import java.util.stream.Collectors;

@Component
public class BoModelBeanLoader extends BeanLoader<IQuery, TenantProject> implements JavaModelBeanLoader {

    @Autowired
    LoaderUtil loaderUtil;

    /**
     * load BoModel 的逻辑：
     * 1、先遍历出所有 BizObject 的子类 得到集合
     * 2、筛选出当中使用 @BizModel 的类，构建 Model，字段包含父类的字段
     * 3、将集合中的剩余类的字段添加到父类对应的Model 中
     *
     * @param project  当前租户
     * @param input    待加载的资源
     * @param consumer 处理加载结果
     * @throws Exception
     */
    @Override
    public void load(TenantProject project, TenantProject input, Consumer<IQuery> consumer) throws Exception {
        TenantClassLoader classLoader = project.getTenantBeanContext().getTenantClassLoader();

        //先扫描到所有的 BizObject (BizObject的所有子类)
        ClassPathScanningCandidateComponentProvider provider = loaderUtil.getScanningBeanProvider();
        DefaultResourceLoader resourceLoader = new DefaultResourceLoader(classLoader);
        provider.setResourceLoader(resourceLoader);
        provider.addIncludeFilter(loaderUtil.getAbstractTypeFilter(BizObject.class));

        //找到所有bo并根据是否有BizModel注解分类
        Map<Boolean, List<BeanDefinition>> isBizModel = provider.findCandidateComponents("logwire.web")
                .stream()
                .collect(Collectors.groupingBy(beanDefinition -> {
                    try {
                        Class<?> clazz = classLoader.loadClass(beanDefinition.getBeanClassName());
                        BizModel bizModelAnnotation = clazz.getAnnotation(BizModel.class);
                        return bizModelAnnotation != null;
                    } catch (ClassNotFoundException e) {
                        throw new ApplicationException("");
                    }
                }));

        //获取需要创建表的bean
        List<BeanDefinition> bizModelBeans = isBizModel.get(true);

        //获取所有不需要创建表的bean(可能为boModel的子类)
        List<BeanDefinition> childrenBeans = isBizModel.get(false);

        //找到拓展字段
        Set<Class<?>> expandoClasses = findExpandoClasses(classLoader,resourceLoader);

        //处理建表boModel,子类boModel,拓展字段之间的关系
        Map<Class, BoModelRelation> boModelRelations = handleBoModel(classLoader, bizModelBeans, childrenBeans, expandoClasses);

        //遍历建表的bean，并找到其拓展字段
        for (BoModelRelation beanDefinition : boModelRelations.values()) {
            Class<?> clazz = beanDefinition.getBoModel();
            //获取类名和注解里的label
            String className = Util.toLowerCaseFirst(clazz.getSimpleName());
            String label = clazz.getAnnotation(BizModel.class).label();
            //独立model
            Model model = initChildModel();
            setName(className,model,label);
            model.setVerboseName(label);
            model.setVerboseNameCode(className);
            model.setIncludeAuditFields(true);
            model.setIncludeVersionField(true);
            model.setIncludeDomainField(true);
            //字段名去重
            Set<String> fieldNames = new HashSet<>();
            //是否没有大文本字段，当一个表有多个大文本字段，只需要建一次表，将多个大文本字段的值放在一张表中
            Boolean isFirstText = true;
            //根据字段建表
            isFirstText = initField(clazz, model, fieldNames,isFirstText, consumer);
            //补充子类
            for (Class childClazz:beanDefinition.getChildrenBo()) {
                isFirstText = initField(childClazz, model, fieldNames,isFirstText, consumer);
            }
            //补充拓展字段
            for (Class expandoClazz:beanDefinition.getExpando()) {
                isFirstText = initField(expandoClazz, model, fieldNames,isFirstText, consumer);
            }
            consumer.accept(model);
        }
    }

    /**
     * 初始化model的字段
     * @param clazz 类
     * @param model model
     * @param consumer
     */
    private Boolean initField(Class<?> clazz, Model model,Set<String> fieldNames,Boolean isFirstText, Consumer<IQuery> consumer) {
        List<Field> fieldList = new ArrayList<>();
        while (clazz != null){
            fieldList.addAll(Arrays.asList(clazz.getDeclaredFields()));
            clazz = clazz.getSuperclass();
        }
        //遍历所有字段
        for (Field f : fieldList) {
            //如果已经加过此字段，则跳过
            if (!fieldNames.add(f.getName()))
                continue;
            if (f.getAnnotation(Column.class) != null) {
                //带有Column注解的字段
                FieldParser parser = FieldParserUtil.getParser(f);
                logwire.core.meta.model.fields.Field field = parser.parse(f.getName(), f);
                model.addField(field);
            } else if (f.getAnnotation(Text.class) != null&&isFirstText) {
                // 带有Text注解的字段 大文本属性会创建一张子表(Model)
                // 子表名为：BO表名_text，一个BO有多个大字段时，使用同一张表
                if (f.getType() != BizText.class) {
                    throw new ApplicationException("@Text 字段类型必须为BizText");
                }
                newTextModel(model, f, consumer);
                isFirstText = false;
            } else if (f.getAnnotation(Array.class) != null) {
                // 带有Array注解的字段 字段类型必须是BizArray<X>, X可以是String/Double/Boolean/Long/Integer/BigDecimal/OffsetDateTime。
                // 数组属性会创建一张子表(Model)，子表名为：BO表名_属性名；包含的字段：
                if (f.getType() != BizArray.class) {
                    throw new ApplicationException("@Array 字段类型必须为BizArray");
                }
                newArrayModel(model, f, consumer);
            } else if (f.getAnnotation(Many.class) != null) {
                // 带有Many或SlaveMany注解的字段 字段类型必须是BizMany<X>, X为BO类型，未指定时可在注解中指定。
                // 主多对多定义会创建一张子表(Model)，子表名为：BO表名_TXT；包含的字段：
                if (f.getType() != BizMany.class) {
                    throw new ApplicationException("@Many 字段类型必须为BizMany");
                }
                newManyModel(model, f, consumer);
            }
        }
        return isFirstText;
    }



    /**
     * 数组字段值存到一个列中，用逗号分隔 字段：
     * owner_id/所属对象ID   (BigInt)
     * value/值     (varchar/numeric/boolean/bigint/integer/timestamp)
     * idx/顺序 Integer
     *
     * @param f
     * @param consumer
     */
    private void newArrayModel(Model model, Field f, Consumer<IQuery> consumer) {
        Array array = f.getAnnotation(Array.class);
        //新建数组字段表
        Model arrayModel = initChildModel();
        arrayModel.setName(model.getName() + "_" + f.getName());
        arrayModel.setVerboseName(array.label());

        ForeignKeyField ownerId = new ForeignKeyField("owner_id", model.getName());
        arrayModel.addField(ownerId);
        arrayModel.addField(new IntegerField("idx"));
        //通过泛型找到字段类型
        Type actualTypeArguments = ((ParameterizedType) f.getGenericType()).getActualTypeArguments()[0];
        FieldParser parser = FieldParserUtil.getParserByFieldType(actualTypeArguments);
        logwire.core.meta.model.fields.Field valueField = parser.parse("value", f);
        arrayModel.addField(valueField);

        consumer.accept(arrayModel);
    }

    /**
     * 大文本 字段：
     * owner_id/所属对象ID   (BigInt)
     * field_name/字段名 varchar(50)
     * value/值     (varchar/numeric/boolean/bigint/integer/timestamp)
     *
     * @param model
     * @param f
     * @param consumer
     */
    private void newTextModel(Model model, Field f, Consumer<IQuery> consumer) {
        Text text = f.getAnnotation(Text.class);
        //新建大文本字段表
        Model textModel = initChildModel();
        textModel.setName(model.getName()+ "_text");
        textModel.setVerboseName(text.label());

        ForeignKeyField ownerId = new ForeignKeyField("owner_id", model.getName());
        textModel.addField(ownerId);
        textModel.addField(new StringField("field_name", 50));
        textModel.addField(new TextField("value"));

        consumer.accept(textModel);
    }

    /**
     * 多对多关联属性 字段：
     * id
     * master_id 主表ID
     * slave_id  从表ID
     *
     * @param model
     * @param f
     * @param consumer
     */
    private void newManyModel(Model model, Field f, Consumer<IQuery> consumer) {
        Many many = f.getAnnotation(Many.class);
        //新建多对多关系表
        Model manyModel = initChildModel();
        manyModel.setName(model.getName() + "_TXT");
        manyModel.setVerboseName(many.label());

        manyModel.addField(new AutoField());
        manyModel.addField(new ForeignKeyField("master_id", model.getName()));
        manyModel.addField(new ForeignKeyField("slave_id", many.referBoName()));
        consumer.accept(manyModel);
    }

    private Model initChildModel(){
        Model childModel = new Model();
        setPrimaryKey(childModel);
        childModel.setIncludeAuditFields(false);
        childModel.setIncludeVersionField(false);
        childModel.setIncludeDomainField(false);
        return childModel;
    }

    /**
     * bo默认主键id为雪花id
     * @param model
     */
    private void setPrimaryKey(Model model){
        BigIntegerField id = new BigIntegerField("id");
        id.setPrimaryKey(true);
        model.addField(id);
    }

    /**
     * 设置表名，如果label不存在，用类名
     * @param className
     * @param model
     * @param label
     */
    private void setName(String className, Model model, String label) {
        if (!Strings.isNullOrEmpty(label)) {
            model.setName(label);
        } else {
            model.setName(className);
        }
    }

    /**
     * 找到拓展字段
     * @param resourceLoader
     * @return
     */
    private Set<Class<?>> findExpandoClasses(TenantClassLoader classLoader,ResourceLoader resourceLoader) {
        ClassPathScanningCandidateComponentProvider provider = loaderUtil.getScanningBeanProvider();
        provider.setResourceLoader(resourceLoader);
        provider.addIncludeFilter(loaderUtil.getInterfaceTypeFilter(BizExpando.class));
        Set<BeanDefinition> candidateComponents = provider.findCandidateComponents("logwire.web");
        return candidateComponents
                .stream()
                .map(e -> {
                    try {
                        return classLoader.loadClass(e.getBeanClassName());
                    } catch (ClassNotFoundException classNotFoundException) {
                        return null;
                    }
                })
                .collect(Collectors.toSet());
    }

    @Override
    public boolean accept(TenantProject input) {
        return true;
    }

    /**
     * 找到model和他的子类和他的拓展字段
     * @param bizModelBeans 需要建表的bo
     * @param childrenBeans bo的子类
     * @param expandoClasses bo的拓展字段
     * @return
     */
    private Map<Class,BoModelRelation> handleBoModel(TenantClassLoader classLoader, List<BeanDefinition> bizModelBeans, List<BeanDefinition> childrenBeans, Set<Class<?>> expandoClasses){
        Map<Class, BoModelRelation> boModel_relation = Maps.newHashMap();

        //遍历所有建表bo，初始化bo关系类
        bizModelBeans.forEach(e->{
            try {
                Class<?> clazz = classLoader.loadClass(e.getBeanClassName());
                boModel_relation.put(clazz,new BoModelRelation(clazz));
            } catch (ClassNotFoundException classNotFoundException) {
            }
        });

        //遍历子类bo
        childrenBeans.forEach(e->{
            try {
                Class<?> childClazz = classLoader.loadClass(e.getBeanClassName());
                boModel_relation.entrySet()
                        .stream()
                        .filter(f->childClazz.isAssignableFrom(f.getKey())||f.getKey().isAssignableFrom(childClazz))
                        .forEach(f->f.getValue().addChildBo(childClazz));
            } catch (ClassNotFoundException classNotFoundException) {
            }
        });

        //遍历所有拓展字段
        expandoClasses.forEach(e->{
            //找到所有接口
            Type[] genericInterfaces = e.getGenericInterfaces();
            for (Type type:genericInterfaces) {
                //找到有泛型的接口
                if (type instanceof ParameterizedType){
                    ParameterizedType paramterizedType = (ParameterizedType) type;
                    //找到BizExpando接口
                    if (paramterizedType.getRawType()==BizExpando.class){
                        //找到泛型
                        Class clazz = (Class)paramterizedType.getActualTypeArguments()[0];
                        //通过泛型去找对应的boRelation
                        Collection<BoModelRelation> relations = boModel_relation.values();
                        for (BoModelRelation relation:relations) {
                            if (relation.checkAndAddExpando(clazz,e))
                                break;
                        }
                        //BizExpando接口只会实现一次，找到后即可跳出循环
                        break;
                    }
                }
            }
        });

        return boModel_relation;
    }

    private class BoModelRelation{

        private final Class boModel;

        private Set<Class> childrenBo = new HashSet<>();

        private Set<Class> expando = new HashSet<>();

        public boolean checkAndAddExpando(Class parameterType,Class expandoClazz){
            if (parameterType.isAssignableFrom(boModel)){
                addExpando(expandoClazz);
                return true;
            }
            for (Class child:childrenBo) {
                if (parameterType.isAssignableFrom(child)){
                    addExpando(expandoClazz);
                    return true;
                }
            }
            return false;
        }

        public BoModelRelation(Class boModel){
            this.boModel = boModel;
        }

        public Class getBoModel() {
            return boModel;
        }

        public Set<Class> getChildrenBo() {
            return childrenBo;
        }

        public Set<Class> getExpando() {
            return expando;
        }

        public void addChildBo(Class clazz){
            childrenBo.add(clazz);
        }

        public void addExpando(Class clazz){
            expando.add(clazz);
        }
    }
}
