package logwire.web.bo.loader;

import com.google.common.collect.Maps;
import logwire.core.bo.annotation.BizModel;
import logwire.core.bo.annotation.ListOperationProvider;
import logwire.core.bo.annotation.ObjectOperationProvider;
import logwire.core.bo.annotation.TypeOperationProvider;
import logwire.core.bo.eventhandler.ListOperationEventHandler;
import logwire.core.bo.eventhandler.ObjectOperationEventHandler;
import logwire.core.bo.eventhandler.OperationEventHandler;
import logwire.core.bo.eventhandler.TypeOperationEventHandler;
import logwire.core.bo.handler.ListOperationHandler;
import logwire.core.bo.handler.ObjectOperationHandler;
import logwire.core.bo.handler.OperationHandler;
import logwire.core.bo.handler.TypeOperationHandler;
import logwire.core.bo.object.BizObject;
import logwire.core.exceptions.ApplicationException;
import logwire.core.resource.BeanLoader;
import logwire.core.tenant.TenantClassLoader;
import logwire.core.tenant.TenantProject;
import logwire.web.bo.BoRelationDefinition;
import logwire.web.bo.proxy.Proxy;
import logwire.web.service.query.QueryService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.ClassPathScanningCandidateComponentProvider;
import org.springframework.context.annotation.ScannedGenericBeanDefinition;
import org.springframework.core.io.DefaultResourceLoader;
import org.springframework.core.type.filter.AnnotationTypeFilter;
import org.springframework.core.type.filter.AssignableTypeFilter;
import org.springframework.stereotype.Component;

import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Consumer;
import java.util.stream.Collectors;

@Component
public class BoRelationBeanDefinitionLoader extends BeanLoader<BoRelationDefinition, TenantProject> {

    private final String BIZ_OBJECT = "biz_object";
    private final String TYPE_OPERATION_PROVIDER = "type_operation_provider";
    private final String OBJECT_OPERATION_PROVIDER = "object_operation_provider";
    private final String LIST_OPERATION_PROVIDER = "list_operation_provider";
    private final String TYPE_OPERATION_HANDLER = "type_operation_handler";
    private final String OBJECT_OPERATION_HANDLER = "object_operation_handler";
    private final String LIST_OPERATION_HANDLER = "list_operation_handler";
    private final String TYPE_OPERATION_EVENT_HANDLER = "type_operation_event_handler";
    private final String OBJECT_OPERATION_EVENT_HANDLER = "object_operation_event_handler";
    private final String LIST_OPERATION_EVENT_HANDLER = "list_operation_event_handler";

    @Autowired
    QueryService queryService;

    @Autowired
    LoaderUtil loaderUtil;

    /**
     * 加载BO相关定义并初始化 service
     *
     * @param project  当前租户
     * @param input    待加载的资源
     * @param consumer 处理加载结果
     * @throws Exception
     */
    @Override
    public void load(TenantProject project, TenantProject input, Consumer<BoRelationDefinition> consumer) throws Exception {
        /*Map<String, BoRelationDefinition> boRelationBeanDefinitions = loadBeanDefinition(project, input);
        boRelationBeanDefinitions.values().forEach(boRelationDefinition -> {
            initializeService(boRelationDefinition, boRelationBeanDefinitions);
            consumer.accept(boRelationDefinition);
        });*/
    }

    /**
     * 初始化
     * 需先初始化 直接继承BizObject的BO的service
     *
     * @param boRelationDefinition
     * @param boRelationBeanDefinitions
     */
    public void initializeService(BoRelationDefinition boRelationDefinition, Map<String, BoRelationDefinition> boRelationBeanDefinitions) {
        if (!boRelationDefinition.isBoServiceInitialized()) {
            Class superClass = boRelationDefinition.getBoClass().getSuperclass();
            if (superClass != BizObject.class) {
                //若非直接继承 先初始化父类
                initializeService(boRelationBeanDefinitions.get(superClass.getSimpleName()), boRelationBeanDefinitions);
            }
            try {
                //create proxy typeOperationProvider
                Object o = Proxy.createTypeOperationProvider(boRelationDefinition);
                Field field = boRelationDefinition.getBoClass().getField("service");
                field.setAccessible(true);
                field.set(null, o);
                boRelationDefinition.setBoServiceInitialized(true);
            } catch (NoSuchFieldException | IllegalAccessException | InstantiationException e) {
                throw ApplicationException.wrap(e);
            }
        }
    }

    /**
     * @param project
     * @return
     * @throws Exception
     */
    private Map<String, BoRelationDefinition> loadBeanDefinition(TenantProject project, TenantProject input) throws Exception {
        input.getBeansOfType(BizObject.class);
        //加载bo类
        Map<String, BoRelationDefinition> boRelationBeanDefinitions = Maps.newHashMap();
        TenantClassLoader classLoader = project.getTenantBeanContext().getTenantClassLoader();
        ClassPathScanningCandidateComponentProvider provider = loaderUtil.getScanningBeanProvider();
        provider.setResourceLoader(new DefaultResourceLoader(classLoader));
        addIncludeFilters(provider);
        // 在 logwire package 下获取所有 bo相关类（bizObject provider handler）
        Set<BeanDefinition> boClassDefinitions = provider.findCandidateComponents("logwire.web");
        Map<String, List<Class<?>>> beanClassMap = getGroupedMap(classLoader, boClassDefinitions);
        registerBeanDefinition(boRelationBeanDefinitions, beanClassMap);
        addOperationProvider(boRelationBeanDefinitions, beanClassMap);
        addOperationHandler(boRelationBeanDefinitions, beanClassMap);
        addOperationEventHandler(boRelationBeanDefinitions, beanClassMap);
        return boRelationBeanDefinitions;
    }

    private Map<String, List<Class<?>>> getGroupedMap(TenantClassLoader classLoader, Set<BeanDefinition> boClassDefinitions) {
        //集合拆分
        return boClassDefinitions
                .stream().map(beanDefinition -> {
                    try {
                        return Class.forName(beanDefinition.getBeanClassName(), true, classLoader);
                    } catch (ClassNotFoundException e) {
                        throw new ApplicationException(e.getMessage(), e);
                    }
                }).collect(Collectors.groupingBy(clazz -> {
                    if (BizObject.class.isAssignableFrom(clazz)) {
                        return BIZ_OBJECT;
                    } else if (TypeOperationProvider.class.isAssignableFrom(clazz)) {
                        return TYPE_OPERATION_PROVIDER;
                    } else if (ObjectOperationProvider.class.isAssignableFrom(clazz)) {
                        return OBJECT_OPERATION_PROVIDER;
                    } else if (ListOperationProvider.class.isAssignableFrom(clazz)) {
                        return LIST_OPERATION_PROVIDER;
                    } else if (TypeOperationHandler.class.isAssignableFrom(clazz)) {
                        return TYPE_OPERATION_HANDLER;
                    } else if (ObjectOperationHandler.class.isAssignableFrom(clazz)) {
                        return OBJECT_OPERATION_HANDLER;
                    } else if (ListOperationHandler.class.isAssignableFrom(clazz)) {
                        return LIST_OPERATION_HANDLER;
                    } else if (TypeOperationEventHandler.class.isAssignableFrom(clazz)) {
                        return TYPE_OPERATION_EVENT_HANDLER;
                    } else if (ObjectOperationEventHandler.class.isAssignableFrom(clazz)) {
                        return OBJECT_OPERATION_EVENT_HANDLER;
                    } else if (ListOperationEventHandler.class.isAssignableFrom(clazz)) {
                        return LIST_OPERATION_EVENT_HANDLER;
                    }
                    return null;
                }));
    }

    private void registerBeanDefinition(Map<String, BoRelationDefinition> boRelationBeanDefinitions, Map<String, List<Class<?>>> beanClassMap) {
        //如果没有使用bo，则不会报错
        if(beanClassMap.get(BIZ_OBJECT)!=null) {
            beanClassMap.get(BIZ_OBJECT).forEach(clazz -> {
                BoRelationDefinition boRelationDefinition = doRegisterBeanDefinition((Class<? extends BizObject>) clazz);
                boRelationBeanDefinitions.put(boRelationDefinition.getId(), boRelationDefinition);
            });
        }
    }

    private void addOperationProvider(Map<String, BoRelationDefinition> boRelationBeanDefinitions, Map<String, List<Class<?>>> beanClassMap) {
        beanClassMap.get(TYPE_OPERATION_PROVIDER).forEach(clazz -> {
            Class<? extends BizObject> type = clazz.getAnnotation(TypeOperationProvider.class).type();
            boRelationBeanDefinitions.get(type.getSimpleName()).addTypeOperationProvider(clazz);
        });
        beanClassMap.get(OBJECT_OPERATION_PROVIDER).forEach(clazz -> {
            Class<? extends BizObject> type = clazz.getAnnotation(ObjectOperationProvider.class).type();
            boRelationBeanDefinitions.get(type.getSimpleName()).addObjectOperationProvider(clazz);
        });
        beanClassMap.get(LIST_OPERATION_PROVIDER).forEach(clazz -> {
            Class<? extends BizObject> type = clazz.getAnnotation(ListOperationProvider.class).type();
            boRelationBeanDefinitions.get(type.getSimpleName()).addListOperationProvider(clazz);
        });
    }

    private void addOperationHandler(Map<String, BoRelationDefinition> boRelationBeanDefinitions, Map<String, List<Class<?>>> beanClassMap) {
        beanClassMap.get(TYPE_OPERATION_HANDLER).forEach(clazz -> {
            try {
                //get BizObject from genericInterface
                BoRelationDefinition boRelationDefinition = getBoRelationDefinitionFromGeneric(boRelationBeanDefinitions, clazz);
                TypeOperationHandler opHandler = (TypeOperationHandler) clazz.newInstance();
                boRelationDefinition.getTypeOperationHandlers().add(opHandler);
            } catch (IllegalAccessException | InstantiationException e) {
                throw ApplicationException.wrap(e);
            }
        });
        beanClassMap.get(OBJECT_OPERATION_HANDLER).forEach(clazz -> {
            try {
                //get BizObject from genericInterface
                BoRelationDefinition boRelationDefinition = getBoRelationDefinitionFromGeneric(boRelationBeanDefinitions, clazz);
                ObjectOperationHandler opHandler = (ObjectOperationHandler) clazz.newInstance();
                boRelationDefinition.getObjectOperationHandlers().add(opHandler);
            } catch (IllegalAccessException | InstantiationException e) {
                throw ApplicationException.wrap(e);
            }
        });
        beanClassMap.get(LIST_OPERATION_HANDLER).forEach(clazz -> {
            try {
                //get BizObject from genericInterface
                BoRelationDefinition boRelationDefinition = getBoRelationDefinitionFromGeneric(boRelationBeanDefinitions, clazz);
                ListOperationHandler opHandler = (ListOperationHandler) clazz.newInstance();
                boRelationDefinition.getListOperationHandlers().add(opHandler);
            } catch (IllegalAccessException | InstantiationException e) {
                throw ApplicationException.wrap(e);
            }
        });
    }

    private void addOperationEventHandler(Map<String, BoRelationDefinition> boRelationBeanDefinitions, Map<String, List<Class<?>>> beanClassMap) {
        beanClassMap.get(TYPE_OPERATION_EVENT_HANDLER).forEach(clazz -> {
            try {
                BoRelationDefinition boRelationDefinition = getBoRelationDefinitionFromGeneric(boRelationBeanDefinitions, clazz);
                List<TypeOperationEventHandler> typeOperationEventHandlers = boRelationDefinition.getTypeOperationEventHandlers();
                TypeOperationEventHandler opHandler = (TypeOperationEventHandler) clazz.newInstance();
                typeOperationEventHandlers.add(opHandler);
            } catch (IllegalAccessException | InstantiationException e) {
                throw ApplicationException.wrap(e);
            }
        });
        beanClassMap.get(OBJECT_OPERATION_EVENT_HANDLER).forEach(clazz -> {
            try {
                //get BizObject from genericInterface
                BoRelationDefinition boRelationDefinition = getBoRelationDefinitionFromGeneric(boRelationBeanDefinitions, clazz);
                ObjectOperationEventHandler opHandler = (ObjectOperationEventHandler) clazz.newInstance();
                boRelationDefinition.getObjectOperationEventHandlers().add(opHandler);
            } catch (IllegalAccessException | InstantiationException e) {
                throw ApplicationException.wrap(e);
            }
        });
        beanClassMap.get(LIST_OPERATION_EVENT_HANDLER).forEach(clazz -> {
            try {
                //get BizObject from genericInterface
                BoRelationDefinition boRelationDefinition = getBoRelationDefinitionFromGeneric(boRelationBeanDefinitions, clazz);
                ListOperationEventHandler opHandler = (ListOperationEventHandler) clazz.newInstance();
                boRelationDefinition.getListOperationEventHandlers().add(opHandler);
            } catch (IllegalAccessException | InstantiationException e) {
                throw ApplicationException.wrap(e);
            }
        });
    }

    private BoRelationDefinition getBoRelationDefinitionFromGeneric(Map<String, BoRelationDefinition> boRelationBeanDefinitions, Class<?> clazz) {
        //get BizObject from genericInterface
        Class<? extends BizObject> bizObject = (Class<? extends BizObject>) ((ParameterizedType) clazz.getGenericInterfaces()[0]).getActualTypeArguments()[0];
        return boRelationBeanDefinitions.get(bizObject.getSimpleName());
    }


    private BoRelationDefinition doRegisterBeanDefinition(Class<? extends BizObject> beanClass) {
        BoRelationDefinition boRelationDefinition = new BoRelationDefinition();
        boRelationDefinition.setId(beanClass.getName());
        boRelationDefinition.setName(beanClass.getName());
        boRelationDefinition.setBoClass(beanClass);
        boRelationDefinition.setBizModel(beanClass.isAnnotationPresent(BizModel.class));
        return boRelationDefinition;
    }

    /**
     * 扫描bo相关类
     * @param provider
     */
    private void addIncludeFilters(ClassPathScanningCandidateComponentProvider provider) {
        provider.addIncludeFilter(loaderUtil.getAbstractTypeFilter(BizObject.class));
        provider.addIncludeFilter(new AssignableTypeFilter(OperationHandler.class));
        provider.addIncludeFilter(new AssignableTypeFilter(OperationEventHandler.class));
        provider.addIncludeFilter(new AnnotationTypeFilter(TypeOperationProvider.class));
        provider.addIncludeFilter(new AnnotationTypeFilter(ObjectOperationProvider.class));
        provider.addIncludeFilter(new AnnotationTypeFilter(ListOperationProvider.class));
    }

    private boolean isConcrete(ScannedGenericBeanDefinition beanDefinition) {
        return beanDefinition.getMetadata().isConcrete();
    }

    @Override
    public boolean accept(TenantProject input) {
        return true;
    }
}
